from ._exp import Experiment
from ._spec import ConcreteImpl, ImplDef, Plugin, SpecDef, impl, spec

__all__ = [
    "ConcreteImpl",
    "Experiment",
    "ImplDef",
    "Plugin",
    "SpecDef",
    "impl",
    "spec",
]



import functools
import inspect
from collections.abc import Callable, Generator, Iterable, Sequence
from typing import Any, Self, overload, override

import attrs
import networkx as nx

from liblaf.grapes.typed import Decorator


def as_seq(value: str | Iterable[str] | None = None) -> Sequence[str]:
    if value is None:
        return ()
    if isinstance(value, (str, bytes)):
        return (value,)
    return tuple(value)


@attrs.frozen
class SpecDef:
    name: str
    firstresult: bool


@attrs.frozen
class ImplDef:
    name: str
    after: str | Sequence[str] = attrs.field(default=(), converter=as_seq)
    before: str | Sequence[str] = attrs.field(default=(), converter=as_seq)
    priority: int = 0


@attrs.frozen
class ConcreteImpl[C: Callable](ImplDef):
    fn: C = attrs.field(kw_only=True)
    plugin_name: str = attrs.field(kw_only=True)


@attrs.define(slots=False)
class Plugin:
    plugins: list[Self] = attrs.field(factory=list)

    @classmethod
    def get_spec(cls, name: str | SpecDef) -> SpecDef:
        if isinstance(name, SpecDef):
            return name
        fn: Callable | None = getattr(cls, name, None)
        if fn is None:
            raise KeyError(name)
        spec: SpecDef | None = getattr(fn, "spec", None)
        if spec is None:
            raise KeyError(name)
        return spec

    @classmethod
    def iter_specs(cls) -> Generator[SpecDef]:
        for _name, fn in inspect.getmembers(cls):
            spec: SpecDef | None = getattr(fn, "spec", None)
            if spec is None:
                continue
            yield spec

    @property
    def name(self) -> str:
        return type(self).__name__

    @functools.cached_property
    def impls(self) -> dict[str, Sequence[ConcreteImpl]]:
        return {spec.name: tuple(self.iter_impls(spec)) for spec in self.iter_specs()}

    def call(self, spec: str | SpecDef, /, *args, **kwargs) -> Any:
        spec: SpecDef = self.get_spec(spec)
        impls: Sequence[ConcreteImpl] = self.impls.get(spec.name, ())
        if spec.firstresult:
            if len(impls) == 0:
                return None
            return impls[0].fn(*args, **kwargs)
        return [impl.fn(*args, **kwargs) for impl in impls]

    def get_impl(self, spec: str | SpecDef, /) -> ConcreteImpl | None:
        spec: SpecDef = self.get_spec(spec)
        fn: Callable | None = getattr(self, spec.name, None)
        if fn is None:
            return None
        impl: ImplDef | None = getattr(fn, "impl", None)
        if impl is None:
            return None
        return ConcreteImpl(
            name=impl.name,
            after=impl.after,
            before=impl.before,
            priority=impl.priority,
            fn=fn,
            plugin_name=self.name,
        )

    def iter_impls(self, spec: str | SpecDef, /) -> Generator[ConcreteImpl]:
        graph = nx.DiGraph()
        impls: dict[str, ConcreteImpl] = {}
        for impl in self.iter_impls_unordered(spec):
            impls[impl.plugin_name] = impl
            graph.add_node(impl.plugin_name)
            for after in impl.after:
                graph.add_edge(after, impl.plugin_name)
            for before in impl.before:
                graph.add_edge(impl.plugin_name, before)
        for impl in nx.lexicographical_topological_sort(
            graph, key=lambda n: impls[n].priority if n in impls else 0
        ):
            if impl not in impls:
                continue
            yield impls[impl]

    def iter_impls_unordered(self, spec: str | SpecDef, /) -> Generator[ConcreteImpl]:
        spec: SpecDef = self.get_spec(spec)
        for plugin in self.plugins:
            impl: ConcreteImpl | None = plugin.get_impl(spec)
            if impl is None:
                continue
            yield impl

    def register(self, plugin: Self) -> None:
        self.plugins.append(plugin)


@overload
def spec(*, firstresult: bool = False) -> Decorator: ...
@overload
def spec[C: Callable](fn: C, /, *, firstresult: bool = False) -> C: ...
def spec(
    fn: Callable | None = None, /, *, firstresult: bool = False
) -> Decorator | Callable:
    if fn is None:
        return functools.partial(spec, firstresult=firstresult)

    @functools.wraps(fn)
    def wrapper(self: Plugin, *args, **kwargs) -> Any:
        return self.call(wrapper.spec, *args, **kwargs)  # pyright: ignore[reportAttributeAccessIssue]

    wrapper.spec = SpecDef(name=fn.__name__, firstresult=firstresult)  # pyright: ignore[reportAttributeAccessIssue]
    return wrapper


@overload
def impl(
    *,
    name: str,
    after: str | Sequence[str] = (),
    before: str | Sequence[str] = (),
    priority: int = 0,
) -> Decorator: ...
@overload
def impl[C: Callable](
    fn: C,
    /,
    *,
    name: str,
    after: str | Sequence[str] = (),
    before: str | Sequence[str] = (),
    priority: int = 0,
) -> C: ...
def impl(
    fn: Callable | None = None,
    /,
    *,
    name: str,
    after: str | Sequence[str] = (),
    before: str | Sequence[str] = (),
    priority: int = 0,
) -> Decorator | Callable:
    if fn is None:
        return functools.partial(
            impl, name=name, after=after, before=before, priority=priority
        )
    fn.impl = ImplDef(name=name, after=after, before=before, priority=priority)  # pyright: ignore[reportFunctionMemberAccess]
    return override(fn)



from typing import override

import attrs

from liblaf.cherries import meta as _info

from . import _abc
from ._exp import exp


@attrs.define
class End(_abc.End):
    priority: int = attrs.field(default=-3, kw_only=True)  # before dvc
    dry_run: bool = False

    @override
    def __call__(self, **kwargs) -> None:
        git_auto_commit(
            "chore(cherries): auto commit (on run end)", dry_run=self.dry_run
        )


@attrs.define
class Start(_abc.Start):
    dry_run: bool = False

    @override
    def __call__(self, **kwargs) -> None:
        git_auto_commit(
            "chore(cherries): auto commit (on run start)", dry_run=self.dry_run
        )


def git_auto_commit(
    header: str = "chore(cherries): auto commit", *, dry_run: bool = False
) -> None:
    body: str = ""
    if exp.exp_name and exp.exp_url:
        body += f"🧪 View experiment {exp.exp_name} at: {exp.exp_url}\n"
    message: str = f"{header}\n\n{body}" if body else header
    _info.git_auto_commit(message, dry_run=dry_run)
    exp.log_other("cherries.git.branch", _info.git_branch())
    exp.log_other("cherries.git.sha", _info.git_commit_sha())
    exp.log_other("cherries.git.url", _info.git_commit_url())



import bisect
import functools
import operator
from collections.abc import Iterable, Mapping
from typing import Any

import attrs
from loguru import logger

from liblaf.cherries.typed import PathLike


@attrs.define
@functools.total_ordering
class Plugin[**P, T]:
    priority: int = attrs.field(default=0, kw_only=True)
    _children: list["Plugin"] = attrs.field(
        factory=list, eq=False, order=False, alias="children"
    )

    def __attrs_post_init__(self) -> None:
        self._children.sort(key=operator.attrgetter("priority"))

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> T:
        ret: T | None = None
        for child in self._children:
            try:
                ret = child(*args, **kwargs)
            except BaseException as err:
                if isinstance(err, (KeyboardInterrupt, SystemExit)):
                    raise
                logger.exception(child)
        return ret  # pyright: ignore[reportReturnType]

    def __lt__(self, other: "Plugin") -> bool:
        if not isinstance(other, Plugin):
            return NotImplemented
        return self.priority < other.priority

    def __eq__(self, other: "Plugin") -> bool:  # pyright: ignore[reportIncompatibleMethodOverride]
        if not isinstance(other, Plugin):
            return NotImplemented
        return self.priority == other.priority

    @property
    def children(self) -> list["Plugin"]:
        return self._children

    def add(self, *children: "Plugin") -> None:
        for c in children:
            bisect.insort(self._children, c, key=operator.attrgetter("priority"))

    def extend(self, children: Iterable["Plugin"]) -> None:
        self.add(*children)

    def remove(self, child: "Plugin") -> None:
        self._children.remove(child)


@attrs.define
class AddTag(Plugin):
    def __call__(self, tag: str, /, **kwargs) -> None:
        super().__call__(tag, **kwargs)


@attrs.define
class AddTags(Plugin):
    def __call__(self, tags: Iterable[str], /, **kwargs) -> None:
        super().__call__(tags, **kwargs)


@attrs.define
class End(Plugin):
    def __call__(self, **kwargs) -> None:
        super().__call__(**kwargs)


@attrs.define
class LogAsset(Plugin):
    def __call__(self, path: PathLike, /, prefix: str | None = None, **kwargs) -> None:
        super().__call__(path, prefix=prefix, **kwargs)


@attrs.define
class LogCode(Plugin):
    def __call__(self, path: PathLike, /, **kwargs) -> None:
        super().__call__(path, **kwargs)


@attrs.define
class LogMetric(Plugin):
    def __call__(
        self,
        key: str,
        value: float,
        /,
        step: int | None = None,
        epoch: int | None = None,
        **kwargs,
    ) -> None:
        super().__call__(key, value, step=step, epoch=epoch, **kwargs)


@attrs.define
class LogMetrics(Plugin):
    def __call__(
        self,
        metrics: Mapping[str, Any],
        /,
        step: int | None = None,
        epoch: int | None = None,
        **kwargs,
    ) -> None:
        super().__call__(metrics, step=step, epoch=epoch, **kwargs)


@attrs.define
class LogOther(Plugin):
    def __call__(self, key: str, value: Any, /, **kwargs) -> None:
        super().__call__(key, value, **kwargs)


@attrs.define
class LogOthers(Plugin):
    def __call__(self, others: Mapping[str, Any], /, **kwargs) -> None:
        super().__call__(others, **kwargs)


@attrs.define
class LogParam(Plugin):
    def __call__(self, name: str, value: Any, /, **kwargs) -> None:
        super().__call__(name, value, **kwargs)


@attrs.define
class LogParams(Plugin):
    def __call__(self, params: Mapping[str, Any], /, **kwargs) -> None:
        super().__call__(params, **kwargs)


@attrs.define
class Start(Plugin):
    def __call__(self, **kwargs) -> None:
        super().__call__(**kwargs)



import datetime
import functools
from pathlib import Path
from typing import Any

from liblaf.cherries import pathutils as _path
from liblaf.cherries.typed import PathLike

from ._spec import Plugin, spec


class Experiment(Plugin):
    @functools.cached_property
    def exp_dir(self) -> Path:
        return _path.exp_dir(absolute=True)

    @property
    def exp_id(self) -> str:
        return self.get_exp_id()

    @property
    def exp_name(self) -> str:
        return self.get_exp_name()

    @exp_name.setter
    def exp_name(self, value: str) -> None:
        self.set_exp_name(value)

    @property
    def exp_url(self) -> str:
        return self.get_exp_url()

    @property
    def project_id(self) -> str:
        return self.get_project_id()

    @property
    def project_name(self) -> str:
        return self.get_project_name()

    @functools.cached_property
    def start_time(self) -> datetime.datetime:
        return datetime.datetime.now().astimezone()

    def log_input(self, path: PathLike, /, **kwargs) -> None:
        self.log_asset(path, prefix="inputs/", **kwargs)

    def log_output(self, path: PathLike, /, **kwargs) -> None:
        self.log_asset(path, prefix="outputs/", **kwargs)

    @spec
    def add_tag(self, tag: str, /) -> None: ...

    @spec
    def add_tags(self, tags: list[str], /) -> None: ...

    @spec
    def end(self) -> None: ...

    @spec(firstresult=True)
    def get_exp_id(self) -> str: ...

    @spec(firstresult=True)
    def get_exp_name(self) -> str: ...

    @spec(firstresult=True)
    def get_exp_url(self) -> str: ...

    @spec(firstresult=True)
    def get_project_id(self) -> str: ...

    @spec(firstresult=True)
    def get_project_name(self) -> str: ...

    @spec(firstresult=True)
    def get_project_url(self) -> str: ...

    @spec
    def log_asset(self, path: PathLike, /, prefix: str | None = None) -> None: ...

    @spec
    def log_code(self, path: str, /) -> None: ...

    @spec
    def log_metric(
        self,
        key: str,
        value: float,
        /,
        step: int | None = None,
        epoch: int | None = None,
    ) -> None: ...

    @spec
    def log_metrics(
        self,
        metrics: dict[str, float],
        /,
        step: int | None = None,
        epoch: int | None = None,
    ) -> None: ...

    @spec
    def log_other(self, key: str, value: Any, /) -> None: ...

    @spec
    def log_others(self, others: dict[str, Any], /) -> None: ...

    @spec
    def log_param(self, key: str, value: Any, /) -> None: ...

    @spec
    def log_params(self, params: dict[str, Any], /) -> None: ...

    @spec
    def set_exp_name(self, name: str, /) -> None: ...

    @spec
    def start(self) -> None: ...



import enum
from collections.abc import Callable, Iterable
from pathlib import Path
from typing import Any

import pydantic

from liblaf import grapes
from liblaf.cherries import pathutils as _path
from liblaf.cherries.typed import PathLike


class AssetKind(enum.StrEnum):
    INPUT = enum.auto()
    OUTPUT = enum.auto()


type PathProvider = (
    PathLike | Iterable[PathLike] | Callable[[PathLike], PathLike | Iterable[PathLike]]
)


class MetaAsset:
    kind: AssetKind
    _extra: PathProvider | None = None

    def __init__(self, kind: AssetKind, extra: PathProvider | None = None) -> None:
        self.kind = kind
        self._extra = extra

    def get_extra(self, value: Path) -> list[Path]:
        if self._extra is None:
            return []
        if callable(self._extra):
            extra: Iterable[PathLike] = grapes.as_iterable(self._extra(value))
            return [Path(p) for p in extra]
        extra: Iterable[PathLike] = grapes.as_iterable(self._extra)
        return [Path(p) for p in extra]


def get_assets(cfg: pydantic.BaseModel, kind: AssetKind) -> list[Path]:
    assets: list[Path] = []
    for name, info in type(cfg).model_fields.items():
        value: Any = getattr(cfg, name)
        if isinstance(value, pydantic.BaseModel):
            assets.extend(get_assets(value, kind))
        for meta in info.metadata:
            if isinstance(meta, MetaAsset) and meta.kind == kind:
                value: Path = Path(value)
                assets.append(value)
                assets.extend(meta.get_extra(value))
    return assets


def get_inputs(cfg: pydantic.BaseModel) -> list[Path]:
    return get_assets(cfg, AssetKind.INPUT)


def get_outputs(cfg: pydantic.BaseModel) -> list[Path]:
    return get_assets(cfg, AssetKind.OUTPUT)


def input(path: PathLike, extra: PathProvider | None = None, **kwargs) -> Path:  # noqa: A001
    field_info: pydantic.fields.FieldInfo = pydantic.Field(_path.data(path), **kwargs)  # pyright: ignore[reportAssignmentType]
    field_info.metadata.append(MetaAsset(kind=AssetKind.INPUT, extra=extra))
    return field_info  # pyright: ignore[reportReturnType]


def output(path: PathLike, extra: PathProvider | None = None, **kwargs) -> Path:
    field_info: pydantic.fields.FieldInfo = pydantic.Field(_path.data(path), **kwargs)  # pyright: ignore[reportAssignmentType]
    field_info.metadata.append(MetaAsset(kind=AssetKind.OUTPUT, extra=extra))
    return field_info  # pyright: ignore[reportReturnType]



from liblaf.cherries import integration


def playground(exp: integration.Experiment) -> integration.Experiment:
    exp.end.add(
        integration.logging.End(),
    )
    exp.start.add(
        integration.logging.Start(),
    )
    return integration.exp
